(* +------------------------------------------------------------------------+
   |                                                                        |
   |                     Suite de Hamming : algorithme AIGU                 |
   |                                                                        |
   +------------------------------------------------------------------------+ *)


open Numerix
module M(E:Int_type) : sig

  (* transforme la famille fournie en famille rduite au sens de Minkowski *)
  (* l'algorithme ne fonctionne qu'en dimension 3                          *)
  val reduction : Linalg.M(E).euclide -> unit

end = struct

  module I = Infixes(E)
  module L = Linalg.M(E)
  open E
  open I
  open L



                        (* +------------------------+
                           |  Ordonne un ttradre  |
                           +------------------------+ *)

  (*
    ABCD est un ttradre dfini par famille (AB,AC,AD)
    Permute les sommets de sorte que ABC soit la face d'inertie minimale
    et A soit oppos au ct mdian sur cette face.
  *)
  let ordonne m =

    (* formules :
	 
       AB^2 = g00       BC^2 = g00 + g11 - 2*g01
       AC^2 = g11       BD^2 = g00 + g22 - 2*g02
       AD^2 = g22       CD^2 = g11 + g22 - 2*g12
	 
       inertie(ABC) = 2*(g00 + g11 - g01)
       inertie(ABD) = 2*(g00 + g22 - g02)
       inertie(ACD) = 2*(g11 + g22 - g12)
       inertie(BCD) = 2*(g00 + g11 + g22 - g01 - g02 - g12)
    *)

    (* change C et D si inertie(ABD) < inertie(ABC) *)
    if m.g.(2).(2) -- m.g.(0).(2) < m.g.(1).(1) -- m.g.(0).(1) then eswap m 1 2;

    (* change B et D si inertie(ACD) < inertie(ABC) *)
    if m.g.(2).(2) -- m.g.(1).(2) < m.g.(0).(0) -- m.g.(0).(1) then eswap m 0 2;

    (* change A et D si inertie(BCD) < inertie(ABC) *)
    if m.g.(2).(2) < m.g.(0).(2) ++ m.g.(1).(2)               then exsub m 2;

    let s = cmp m.g.(1).(1) m.g.(0).(0) in

    (* change A et B si mdian = AC *)
    if s * (cmp m.g.(0).(0) (m.g.(0).(1)*.2)) > 0          then exsub m 0

    (* change A et C si mdian = AB *)
    else if s * (cmp m.g.(1).(1) (m.g.(0).(1)*.2)) < 0     then exsub m 1


                         (* +-----------------------+
                            |  algorithme de Gauss  |
                            +-----------------------+ *)

  (*
    transforme (v0,v1) en une base minimale du groupe engendr.
    Le triangle form par cette base minimale a tous ses angles aigus.
  *)
  let gauss m =

    (* ici on suppose que |v0| <= |v1|, rduit v1 si besoin *)
    let rec loop() =
      let q,r = quomodc m.g.(0).(1) m.g.(0).(0) in
      if q <> zero then begin
	epsub  m 1 0 q;
	if m.g.(1).(1) < m.g.(0).(0) then begin eswap m 0 1; loop() end
      end

    in
    if m.g.(1).(1) < m.g.(0).(0) then eswap m 0 1;
    loop();

    (* si l'angle (v0,v1) est obtus alors change le signe de v1 *)
    if m.g.(0).(1) < zero then eneg m 1


                        (* +-------------------------+
                           |  Redresse un ttradre  |
                           +-------------------------+ *)
      
  (*
    Dplace v2 de sorte qu'il se projette  l'intrieur du triangle (v0,v1).
  *)
  let redresse m =

    (* projection orthogonale de D sur le plan (u,v) = (xu+yv)/z *)
    let uu = m.g.(0).(0)
    and uv = m.g.(0).(1)
    and uw = m.g.(0).(2)
    and vv = m.g.(1).(1)
    and vw = m.g.(1).(2)    in
    let x = uw**vv -- vw**uv
    and y = uu**vw -- uv**uw
    and z = uu**vv -- sqr(uv) in

    (* arrondit et teste la position par rapport  la droite (uv) *)
    let p,r = quomod x z
    and q,s = quomod y z in
    if r++s <= z then begin
      epsub m 2 0 p;
      epsub m 2 1 q;
    end
    else begin
      let p = p+.1 and q = q+.1 in
      epsub m 2 0 p;
      epsub m 2 1 q;
      eneg m 2
    end
	

                           (* +-------------------+
                              |  Algorithme Aigu  |
                              +-------------------+ *)

  (* construit un ttradre suraigu  partir de (v0,v1,v2) *)
  let rec aigu m =

    (* rduit la face d'inertie minimale et rapproche le dernier sommet *)
    ordonne  m;
    gauss    m;
    redresse m;

    (* angle ADB obtus ? *)
    if m.g.(0).(2) > m.g.(2).(2) then begin
      (* symtrise C % milieu(A,B) *)
      if m.g.(0).(0) ++ m.g.(1).(2)*.4 < (m.g.(0).(1) ++ m.g.(0).(2))*.2
      then esub m 2 1;
      aigu(m)
    end

    (* angle ADC obtus ? *)
    else if m.g.(1).(2) > m.g.(2).(2) then begin
      (* symtrise B % milieu(A,C) *)
      if m.g.(1).(1) ++ m.g.(0).(2)*.4 < (m.g.(1).(0) ++ m.g.(1).(2))*.2
      then esub m 1 2;
      aigu(m)
    end

    (* angle BDC obtus ? *)
    else if m.g.(0).(2) ++ m.g.(1).(2) > m.g.(0).(1) ++ m.g.(2).(2) then begin
      (* symtrise A % milieu(B,C) *)
      if m.g.(0).(0) ++ m.g.(1).(1) < (m.g.(0).(2) ++ m.g.(1).(2) -- m.g.(0).(1))*.2
      then esub2 m 2 0 1;
      aigu(m)
    end
      

                            (* +-----------------+
                               |  Base minimale  |
                               +-----------------+ *)

  let reduction m =

    if Array.length(m.g) <> 3 then failwith "Algorithme Aigu en dimension <> 3";

    aigu(m);
    let u2 = m.g.(0).(0)
    and v2 = m.g.(1).(1)
    and w2 = m.g.(2).(2)
    and uv = m.g.(0).(1)
    and uw = m.g.(0).(2)
    and vw = m.g.(1).(2) in


    (* liste des artes du ttradre avec longueurs *)
    let l = ref [
      ([| 1; 0; 0|], u2);              (* AB *)
      ([| 0; 1; 0|], v2);              (* AC *)
      ([| 0; 0; 1|], w2);              (* AD *)
      ([|-1; 1; 0|], u2 ++ v2 -- uv*.2); (* BC *)
      ([|-1; 0; 1|], u2 ++ w2 -- uw*.2); (* BD *)
      ([| 0;-1; 1|], v2 ++ w2 -- vw*.2)  (* CD *)
    ] in

    (* arte AD obtuse => symtrise % milieu(AD) *)
    if w2**uv < uw**vw then
      l := [
	([|-1;-1; 1|], u2 ++ v2 ++ w2 ++ (uv -- uw -- vw)*.2); (* BC' *)
	([|-2; 0; 1|], (u2 -- uw)*.4 ++ w2);                (* BB' *)
	([| 0;-2; 1|], (v2 -- vw)*.4 ++ w2)                 (* CC' *)
      ] @ !l
      
    (* arte BD obtuse => symtrise % milieu(BD) *)
    else if (u2--uw)**(v2--vw) < (uw--uv)**(uv--vw) then
      l := [
	([| 1;-1; 1|], u2 ++ v2 ++ w2 ++ (uw -- uv -- vw)*.2);   (* AC' *)
	([| 0; 1; 1|], v2 ++ w2 ++ vw*.2);                    (* AA' *)
	([| 1;-2; 1|], u2 ++ w2 ++ uw*.2 ++ (v2 -- uv -- vw)*.4) (* CC' *)
      ] @ !l

    (* arte CD obtuse => symtrise % milieu(CD) *)
    else if (v2--vw)**(u2--uw) < (vw--uv)**(uv--uw) then
      l := [
	([|-1; 1; 1|], v2 ++ u2 ++ w2 ++ (vw -- uv -- uw)*.2);   (* AB' *)
	([| 1; 0; 1|], u2 ++ w2 ++ uw*.2);                    (* AA' *)
	([|-2; 1; 1|], v2 ++ w2 ++ vw*.2 ++ (u2 -- uv -- uw)*.4) (* BB' *)
      ] @ !l;

    (* classe les artes par longueur croissante *)
    l := Sort.list (fun (_,x) (_,y) -> x < y) !l;

    (* les deux premiers minimums,x,y, sont linairement indpendants *)
    (* les extrait et cherche le premier vecteur indpendant de x,y   *)
    let p = match !l with
      | (x,_)::(y,_)::suite ->
	  let a = x.(1)*y.(2) - x.(2)*y.(1)
	  and b = x.(2)*y.(0) - x.(0)*y.(2)
	  and c = x.(0)*y.(1) - x.(1)*y.(0) in
	  let rec get = function
	    | [] -> failwith "cas impossible"
	    | (z,_)::reste -> if a*z.(0) + b*z.(1) + c*z.(2) <> 0
	      then [|x;y;z|] else get reste
	  in get suite
      | _ -> failwith "cas impossible"
    in

    (* forme la nouvelle matrice *)
    echg m (Array.map (Array.map of_int) p)

end
