(* +------------------------------------------------------------------------+
   |                                                                        |
   |               Suite de Hamming : Dcomposition d'un cne               |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

open Printf
open Numerix

module M(E:Int_type) : sig

  (* cne avec coefficient et lments prcalculs                       *)
  (* les vecteurs sont ordonns de sorte que le dterminant soit positif *)
  (* ceci ne marche pas en dimension 1                                   *)
  type cone = {s : int;                   (* coefficient                 *)
	       f : Linalg.M(E).matrice;   (* vecteurs supportant le cne *)
               g : Linalg.M(E).matrice;   (* comatrice                   *)
               d : E.t}                   (* dterminant du cne         *)

  val make  : int -> Linalg.M(E).matrice -> cone (* cone(s,f) *)
  val dual  : cone -> cone                       (* cne dual *)
  val print : out_channel -> cone -> unit        (* affichage *)

  (* Dcompose une liste de cnes en combinaison  coefficients +/-1 *)
  (* de cnes unimodulaires en ngligeant les cnes de dimension < n *)
  (* le premier paramtre est la fonction de rduction               *)
  val barvinok : (Linalg.M(E).euclide -> unit) -> cone list -> cone list

end = struct

  module I = Infixes(E)
  module L = Linalg.M(E)
  open E
  open I
  open L
  let un    = of_int(1)

  (* cne avec coefficient et lments prcalculs *)
  type cone = {s : int;       (* coefficient                 *)
	       f : matrice;   (* vecteurs supportant le cne *)
               g : matrice;   (* comatrice                   *)
               d : E.t}       (* dterminant du cne         *)

  (* affiche un cne *)
  let print ch cone =
    fprintf ch "cone{ s=%d;\n" cone.s;
    fprintf ch "f=\n"; fprint ch cone.f;
    fprintf ch "g=\n"; fprint ch cone.g;
    fprintf ch "d=%s}\n" (string_of cone.d)

  (* cne construit sur une famille *)
  let make s f =
    let f = Array.map Array.copy f in
    let g = comatrice(f)           in
    let d = ps f.(0) g.(0)         in
    if d < zero then begin
      fswap f 0 1;
      fswap g 0 1;
      Array.iter vneg_in g
    end;
    {s=s; f=f; g=g; d=abs(d)}
      
  (* cne dual *)
  let dual {s=s; f=f; g=g; d=d} =
    let n   = Array.length f           in
    let dn2 = pow d (n-2)             in (* com(com f) = det(f)^(n-2)*f *)
    let ccf = Array.map (( *** ) dn2) f in
    let dn1 = d**dn2                    in (* det(com f) = det(f)^(n-1)   *)
    {s=s; f=g; g=ccf; d=dn1}

                      (* +-----------------------------+
                         |  Dcomposition de Barvinok  |
                         +-----------------------------+ *)

  let rec barvinok reduction = function
  | [] -> []
  | ({s=s;f=f;g=g;d=d} as cone)::suite ->

      if      d = zero then barvinok reduction suite          (* cne ngligeable  *)
      else if d = one   then cone :: barvinok reduction suite  (* cne unimodulaire *)
      else begin                                              (* cne de dterminant > 1 *)

	(* redresse le simplexe *)
	let n = Array.length(f) in
        let base = {L.v = id(n); L.g = gram(transpose g)} in
	reduction(base);

	(* passe en revue tous les petits vecteurs jusqu' en trouver      *)
	(* un tel que tous les dterminants soient <= d^((n-1)/n)          *)
	let d' = root (pow d (n-1)) n in (* borne pour les dterminants   *)
	let x0   = Array.make n zero   in (* coord. dans la base rduite   *)
	let x    = Array.make n zero   in (* coord. dans la base canonique *)
	let dets = Array.make n zero   in (* dterminants                  *)
        let trouv = ref(false)        in

	while not(!trouv) do
	  vnext(x0);
          vcomb_in x x0 base.v;
	  let i = ref(0) in
	  trouv := true;
	  while (!i < n) && !trouv do
	    dets.(!i) <- ps g.(!i) x;
	    trouv := abs(dets.(!i)) <= d';
	    incr i
	  done
	done;

	(* change le signe de x pour avoir au moins un dterminant positif *)
	let i = ref(0) in
	while (!i < n) & (dets.(!i) >= zero) do incr i done;
	if !i < n then begin vneg_in x; vneg_in dets end;

	(* empile les petits cnes non singuliers *)
	let l = ref(suite) in
	for i=0 to n-1 do match sgn(dets.(i)) with
	  | 0 -> ()
	  | 1 -> let fi = Array.map Array.copy f in
                 fi.(i) <- Array.copy x;
	         l := {s=s; f=fi; g=comatrice(fi); d=dets.(i)} :: !l
	  | _ -> let fi = Array.map Array.copy f in
                 fi.(i) <- Array.copy x;
	         fswap fi 0 1;
	         l := {s = -s; f=fi; g=comatrice(fi); d=abs(dets.(i))} :: !l
	done;

	(* termine rcursivement *)
	barvinok reduction !l

      end

end
