(* +-------------------------------------------------------------------------+
   |                                                                         |
   |    Suite de Hamming, mthode volumtrique avec l'algorithme de Barvinok |
   |                                                                         |
   +-------------------------------------------------------------------------+ *)

open Printf
open Numerix

module Main(E:Int_type) = struct

  module I = Infixes(E)
  open E
  open I

  module Logs     = Logs.M(E)
  module Linalg   = Linalg.M(E)
  module LLL      = Lll.M(E)
  module Aigu     = Aigu.M(E)
  module Simplexe = Simplexe.M(E)
  open Linalg

  let zero  = of_int(0)
  let un    = of_int(1)
  let deux  = of_int(2)
  let trois = of_int(3)
  let cinq  = of_int(5)

  let debug = ref(false) (* affichage des tapes du calcul   *)
  exception Echec        (* en cas de prcision insuffisante *)

          (* +------------------------------------------+
             |  calcule ln(2), ln(3), ln(5) sur n bits  |
             |  retourne les approximations par dfaut, |
             |  par excs, les approximations rduites  |
             |  et le ttradre associ.                |
             +------------------------------------------+ *)

  let ttra reduction p =
      let li2,li3,li5,ls2,ls3,ls5 = Logs.ln235 p in
      if !debug then begin
	chrono(sprintf "2^%d*ln(2) = %s + %s" p (string_of li2) (string_of(ls2--li2)));
	chrono(sprintf "2^%d*ln(3) = %s + %s" p (string_of li3) (string_of(ls3--li3)));
	chrono(sprintf "2^%d*ln(5) = %s + %s" p (string_of li5) (string_of(ls5--li5)));
      end;

      let fam = {v=id(3); g=gram([|[|un;   li3;     li5    |];
                                   [|zero; neg li2; zero   |];
                                   [|zero; zero;    neg li2|]|])} in
      reduction fam;    
      let a = abs(fam.v.(0).(0))
      and b = abs(fam.v.(0).(1))
      and c = abs(fam.v.(0).(2)) in
      if !debug then begin
	chrono (sprintf "Rduction a = %s" (string_of a));
	chrono (sprintf "          b = %s" (string_of b));
	chrono (sprintf "          c = %s" (string_of c));
      end;

      let s = Simplexe.make reduction [| a; b; c |] in
      if !debug then chrono (sprintf "Barvinok -> %d cnes"
			       (Simplexe.nbcones s));
      (li2,li3,li5,ls2,ls3,ls5,a,b,c,s)

      (* +--------------------------------------------------+
         |  calcule le n-me entier de Hamming avec p bits  |
         +--------------------------------------------------+ *)

  let hamming reduction n p =      
    let (li2,li3,li5,ls2,ls3,ls5,a,b,c,s) = ttra reduction p in
      
    (* Rsolution de approx(a,b,c,k) = n+1 par la mthode de Newton *)
    (* arrte  +/- 2                                               *)
    let rec newton k =
      let v,dv,dk = Simplexe.approx s k in
      if !debug then chrono(sprintf "Approx k = %s v ~ %s"
			      (string_of k) (string_of v));
      let t = v -- n -- un  in
      let k = k -- t**dk//dv in
      if abs(t) < trois then k,dv,dk else newton k
    in
    let k,dv,dk = newton (root (a**b**c**n*.6) 3) in

    (* k0 = plus petit k connu tel que volume(k) = n   *)
    (* k1 = plus grand k connu tel que volume(k) = n+1 *)
    (* k2 = plus grand k connu tel que volume(k) <= n  *)
    (* k3 = plus petit k connu tel que volume(k) > n   *)

    let k0 = ref(None) and k1 = ref(None)
    and k2 = ref(None) and k3 = ref(None) in

    (* mise  jour de k0..k3 aprs un calcul exact de volume *)
    let kmin kref k = match !kref with
      | None    -> kref := Some(k)
      | Some(t) -> if t > k then kref := Some(k)
    and kmax kref k = match !kref with
      | None    -> kref := Some(k)
      | Some(t) -> if t < k then kref := Some(k)
    in

    let calcul(k) =
      let v = Simplexe.volume s k in
      if !debug then chrono (sprintf "Exact  k = %s v = %s"
			       (string_of k) (string_of v));
      let t = v -- n in
      if t < zero      then       kmax k2 k
      else if t = zero then begin kmax k2 k; kmin k0 k end
      else if t = un   then begin kmin k3 k; kmax k1 k end
      else                        kmin k3 k;
      v
    in

    (* encadre le n-me point de Hamming par approximations successives *)
    let rec monte k v = if !k3 == None then begin
      let k' = k -- (v--n--un)**dk//dv in
      let v' = calcul(k')         in
      monte k' v'
    end

    and descend k v = if !k2 == None then begin
      let k' = k ++ (n--v)**dk//dv in
      let v' = calcul(k')      in
      descend k' v'
    end

    and dicho() = if (!k0 == None) or (!k1 == None) then begin
      match !k2,!k3 with
	| Some(a),Some(b) ->
	    let k = (a++b)/.2 in
	    if k > a then let _ = calcul(k) in dicho() else raise Echec
	| _ -> failwith "cas impossible"
    end
    in

    let v = calcul(k) in
    monte k v;
    descend k v;
    dicho();
    let k,k' = match !k1,!k0 with
      | Some(a),Some(b) -> a,b
      | _ -> failwith "cas impossible"
    in

    (* l'identifie par calcul de drives partielles *)
    let l = Simplexe.derive s k k' in
    let x = l.(0) and y = l.(1) and z = l.(2) in
    if !debug then chrono(sprintf "Drive -> x = %s y = %s z = %s"
			    (string_of x) (string_of y) (string_of z));

    (* validation *)
    let ni = x**li2 ++ y**li3 ++ z**li5
    and ns = x**ls2 ++ y**ls3 ++ z**ls5 in
      
    let check d li ls =
      let ki = (d**ni)//ls and ks = (d**ns ++ li -- un)//li in
      if (match !k0 with
	    | None -> failwith "cas impossible"
	    | Some(k) -> (k > ki) & (calcul(ki) < n)
	 ) or (match !k1 with
            | None -> failwith "cas impossible"
	    | Some(k) -> (k < ks) & (calcul(ks) > n++un)
	 ) then raise Echec
      in

      check a li2 ls2;
      check b li3 ls3;
      check c li5 ls5;
      if !debug then chrono("Validation -> OK")
      else printf "%s %s %s\n" (string_of x) (string_of y) (string_of z)


    (* +------------------------------------------------------+
       |  Calcule le rang d'un entier de Hamming avec p bits  |
       +------------------------------------------------------+ *)

  let rang reduction x y z p =
    let (li2,li3,li5,ls2,ls3,ls5,a,b,c,s) = ttra reduction p in

    (* encadre le triangle de Hamming passant par (x,y,z) *)
    let ni = x**li2 ++ y**li3 ++ z**li5
    and ns = x**ls2 ++ y**ls3 ++ z**ls5 in
    let k  = min ((a**ni)//ls2) (min ((b**ni)//ls3) ((c**ni)//ls5))
    and k' = max ((a**ns)//li2) (max ((b**ns)//li3) ((c**ns)//li5)) in

    (* calcule les volumes et vrifie qu'ils diffrent de 1 *)
    let v  = Simplexe.volume s k  in
    if !debug then chrono(sprintf "Volume k  = %s v  = %s"
			    (string_of k) (string_of v));
    let v' = Simplexe.volume s k' in
    if !debug then chrono(sprintf "Volume k' = %s v' = %s"
			    (string_of k') (string_of v'));

    if v'--v > un then raise Echec
    else if not !debug then printf "%s\n" (string_of v)


                    (* +-----------------------+
                       |  Programme principal  |
                       +-----------------------+ *)

  module Parser = Parser.M(E)
  open Parser
  type action = Hamming of E.t | Rang of E.t*E.t*E.t
  type reduction = Aigu | LLL

  let main arglist =

    (* analyse la ligne de commande *)
    let rec parse (action,reduc,p) = function
      | "-d"::s          -> debug := true; parse (action,reduc,p) s
      | "-aigu"::s       -> parse (action,Aigu,p) s
      | "-lll"::s        -> parse (action,LLL,p) s
      | "-p"::p::s       -> parse (action,reduc,int_of_string(p)) s
      | "-r"::x::y::z::s -> parse (Rang(evalue x, evalue y, evalue z),reduc,p) s
      | "-n"::n::s       -> parse (Hamming(evalue n),reduc,p) s
      | n::s             -> parse (Hamming(try evalue n with _ -> raise Exit),reduc,p) s
      | []               -> (action,reduc,p)
    in
    let (action,reduc,p) = try parse (Hamming(zero),LLL,10) (List.tl arglist)
    with
      | Incorrect(s) ->
          fprintf stderr "expression incorrecte : %s\n" s;
          flush stderr;
	  exit(1)

      | _ ->
	  fprintf stderr "syntaxe : hamming [-e entiers] [-d] [-p p] [-aigu | -lll] ([-n] n | -r x y z)\n";
          flush stderr;
	  exit(1)
    in

    (* contrle les paramtres et fixe la prcision initiale *)
    let p = ref(match action with
      | Hamming(n) ->
	  if n >= zero then p + nbits(n)
	  else begin
	    fprintf stderr "erreur : n < 0\n";
	    flush stderr;
	    exit(1)
	  end
      | Rang(x,y,z) ->
	  if (x >= zero) & (y >= zero) & (z >= zero)
	  then p + nbits(x) + nbits(y) + nbits(z)
	  else begin
	    if x < zero then fprintf stderr "erreur : x < 0\n";
	    if y < zero then fprintf stderr "erreur : y < 0\n";
	    if z < zero then fprintf stderr "erreur : z < 0\n";
	    flush stderr;
	    exit(1)
	  end)
    in

    (* algorithme de rduction *)
    let reduction = match reduc with
      | Aigu -> Aigu.reduction
      | LLL  -> LLL.reduction
    in

    (* effectue les calculs avec p bits de rserve *)
    (* recommence en cas d'chec en augmentant p   *)
    if !debug then begin
      chrono (sprintf "Entiers   : %s" (name()));
      chrono(sprintf  "Rduction : %s"
	       (match reduc with Aigu -> "Aigu" | LLL -> "LLL"))
    end;
    while try match action with
      | Hamming(n)  -> hamming reduction n !p;  false
      | Rang(x,y,z) -> rang reduction x y z !p; false
    with Echec -> true
    do if !debug then chrono "Echec"; p := !p + 8 done;

    flush stderr;
    flush stdout

end (* Main *)


(* rglage du GC : tas mineur = 1M mots *)
let _ = Gc.set {(Gc.get()) with Gc.minor_heap_size = 1000000}

module S = Start(Main)
let _ = S.start()

