(* +------------------------------------------------------------------------+
   |                                                                        |
   |        Suite de Hamming, oprations sur les familles de vecteurs       |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

open Numerix
module M(E:Int_type) : sig

  (*
    un vecteur est un tableau de coordonnes entires
    une famille de vecteurs est un tableau de vecteurs
    une famille euclidienne est un couple (famille, matrice de Gram)
  *)

  type vecteur = E.t array
  type matrice = vecteur array
  type euclide = {v:matrice; g:matrice}

  (* affichage *)
  val vprint : out_channel -> vecteur -> unit
  val fprint : out_channel -> matrice -> unit
  val eprint : out_channel -> euclide -> unit

  (* oprations hors place sur les vecteurs *)
  (*            x            u          v   *)
  val ( +++ ):            vecteur -> vecteur -> vecteur     (* u + v *)
  val ( --- ):            vecteur -> vecteur -> vecteur     (* u - v *)
  val ( *** ): E.t ->     vecteur -> vecteur                (* x*u   *)
  val vneg   :            vecteur -> vecteur                (* -u    *)
  val vcomb  : vecteur -> matrice -> vecteur                (* sum(x.(i)*u.(i))        *) 
  val ps     :            vecteur -> vecteur -> E.t         (* produit scalaire        *)
  val pv     :            vecteur -> vecteur -> vecteur     (* p. vect. en dimension 3 *)

  (* oprations sur place sur les vecteurs *)
  (*               u          x/v       w  *)
  val vsub_in  : vecteur -> vecteur ->            unit  (* u <- v - u     *)
  val vpsub_in : vecteur -> E.t     -> vecteur -> unit  (* u <- u - xw    *)
  val vneg_in  : vecteur ->                       unit  (* u <- -u        *)
  val vsym2_in : vecteur -> vecteur -> vecteur -> unit  (* u <- v + w - u *)
  val vcomb_in : vecteur -> vecteur -> matrice -> unit  (* u <- sum(x.(i)*w.(i) *) 

  (* oprations sur place sur les familles de vecteurs *)
  (*            v         i        j                   *)
  val fswap : matrice -> int   -> int      -> unit   (* vi <-> vj *)

  (* oprations sur place sur les familles euclidiennes *)
  (*             v        i      j      p/k             *)
  val esub  : euclide -> int -> int ->        unit   (* vi <- vj - vi  *)
  val epsub : euclide -> int -> int -> E.t -> unit   (* vi <- vi - pvj *)
  val eneg  : euclide -> int ->               unit   (* vi <- -vi      *)
  val esub2 : euclide -> int -> int -> int -> unit   (* vi <- vj+vk-vi *)
  val eswap : euclide -> int -> int ->        unit   (* vi <-> vj      *)

  (* xsub v i remplace tous les vj par vj-vi et remplace vi par -vi *)
  (* = changement d'origine dans un simplexe *)
  val fxsub : matrice -> int -> unit
  val exsub : euclide -> int -> unit

  (* chg v m applique  v le changement de base dfini par la matrice m *)
  val fchg  : matrice -> matrice -> unit
  val echg  : euclide -> matrice -> unit

  (* oprations diverses *)
  val transpose : matrice -> matrice  (* transposition    *)
  val det       : matrice -> E.t      (* dterminant      *)
  val comatrice : matrice -> matrice  (* comatrice        *)
  val gram      : matrice -> matrice  (* matrice de Gram  *)
  val id        : int     -> matrice  (* matrice identit *)
  val vnext     : vecteur -> unit     (* numration des vecteurs primitifs *)

  val quomodc   : E.t -> E.t -> E.t*E.t (* division centre *)

end =  struct

  type vecteur = E.t array
  type matrice = vecteur array
  type euclide = {v:matrice; g:matrice}

  module I = Infixes(E)
  open E
  open I
  open Printf

                         (* +-------------+
                            |  Affichage  |
                            +-------------+ *)

  (* affiche un vecteur *)
  let vprint ch v =
    fprintf ch "[";
    for i=0 to Array.length(v)-1 do
      if i<>0 then fprintf ch ",";
      fprintf ch "%s" (string_of v.(i))
    done;
    fprintf ch "]\n"

  (* affiche une famille *)
  let fprint ch f =
    let m = Array.map (Array.map string_of) f in
    let l = Array.fold_left
	      (Array.fold_left (fun a b -> Pervasives.max a (String.length b)))
	      0 m
    in
    for i=0 to Array.length(m)-1 do
      fprintf ch "[";
      for j=0 to Array.length(m.(i))-1 do
	if j<>0 then fprintf ch ",";
	for k = String.length(m.(i).(j)) to l do fprintf ch " " done;
	fprintf ch "%s" m.(i).(j)
      done;
      fprintf ch "]\n"
    done;
    fprintf ch "\n"

  (* affiche une famille euclidienne *)
  let eprint ch {v=v; g=g} = fprint ch v; fprint ch g

          (* +------------------------------------------+
             |  oprations hors place sur les vecteurs  |
             +------------------------------------------+ *)

  (* addition *)
  let ( +++ ) u v =
    let n = Array.length(u)   in
    let w = Array.make n zero in
    for i=0 to n-1 do w.(i) <- u.(i) ++ v.(i) done;
    w

  (* soustraction *)
  let ( --- ) u v =
    let n = Array.length(u)   in
    let w = Array.make n zero in
    for i=0 to n-1 do w.(i) <- u.(i) -- v.(i) done;
    w

  (* produit externe *)
  let ( *** ) a u =
    let n = Array.length(u)   in
    let v = Array.make n zero in
    for i=0 to n-1 do v.(i) <- a**u.(i) done;
    v

  (* oppos *)
  let vneg u =
    let n = Array.length(u)   in
    let v = Array.make n zero in
    for i=0 to n-1 do v.(i) <- neg u.(i) done;
    v


  (* combinaison linaire *)
  let vcomb c v =
    let n = Array.length(v.(0)) in
    let w = Array.make n zero   in
    for i = 0 to Array.length(c)-1 do for j = 0 to n-1 do
      w.(j) <- w.(j) ++ c.(i)**v.(i).(j)
    done done;
    w

  (* produit scalaire *)
  let ps u v =
    let r = ref(zero) in
    for k=0 to Array.length(u)-1 do r := !r ++ u.(k)**v.(k) done;
    !r

  (* produit vectoriel, limit  la dimension 3 *)
  let pv u v = [| u.(1)**v.(2) -- u.(2)**v.(1);
                  u.(2)**v.(0) -- u.(0)**v.(2);
                  u.(0)**v.(1) -- u.(1)**v.(0) |]


           (* +-----------------------------------------+
              |  oprations sur place sur les vecteurs  |
              +-----------------------------------------+ *)

  (* u <- v - u *)
  let vsub_in u v =
    for k=0 to Array.length(u)-1 do u.(k) <- v.(k) -- u.(k) done

  (* u <- u - pv *)
  let vpsub_in u x w =
      for k=0 to Array.length(u)-1 do u.(k) <- u.(k) -- x**w.(k) done

  (* u <- -u *)
  let vneg_in u =
    for k=0 to Array.length(u)-1 do u.(k) <- neg u.(k) done

  (* u <- v + w - u *)
  let vsym2_in u v w =
    for k=0 to Array.length(u)-1 do u.(k) <- v.(k) ++ w.(k) -- u.(k) done
      
  (* combinaison linaire *)
  let vcomb_in u x w =
    let n = Array.length(u) in
    for i=0 to n-1 do
      let s = ref(zero) in
      for j=0 to Array.length(w)-1 do s := !s ++ x.(j)**w.(j).(i) done;
      u.(i) <- !s
    done


     (* +-----------------------------------------------------+
        |  oprations sur place sur les familles de vecteurs  |
        +-----------------------------------------------------+ *)

  (* vi <-> vj *)
  let fswap v i j =
    let x = v.(i) in v.(i) <- v.(j); v.(j) <- x

    
    (* +------------------------------------------------------+
       |  oprations sur place sur les familles euclidiennes  |
       +------------------------------------------------------+ *)

  (* vi <- vi - pvj *)
  let epsub {v=v; g=g} i j p =
    vpsub_in v.(i) p v.(j);
    vpsub_in g.(i) p g.(j);
    for k=0 to Array.length(g)-1 do g.(k).(i) <- g.(i).(k) done;
    g.(i).(i) <- g.(i).(i) -- p**g.(j).(i)

  (* vi <- -vi *)
  let eneg {v=v; g=g} i =
    vneg_in v.(i);
    vneg_in g.(i);
    for k=0 to Array.length(g)-1 do g.(k).(i) <- g.(i).(k) done;
    g.(i).(i) <- neg g.(i).(i)

  (* vi <- vj - vi  *)
  let esub {v=v; g=g} i j =
    vsub_in v.(i) v.(j);
    vsub_in g.(i) g.(j);
    for k=0 to Array.length(g)-1 do g.(k).(i) <- g.(k).(j) -- g.(k).(i) done

  (* vi <- vj+vk-vi *)
  let esub2 {v=v; g=g} i j k =
    vsym2_in v.(i) v.(j) v.(k);
    vsym2_in g.(i) g.(j) g.(k);
    for l=0 to Array.length(g)-1 do
      g.(l).(i) <- g.(l).(j) ++ g.(l).(k) -- g.(l).(i)
    done
      
  (* vi <-> vj      *)
  let eswap {v=v; g=g} i j =
    fswap v i j;
    fswap g i j;
    for k=0 to Array.length(g)-1 do
      let x = g.(k).(i) in g.(k).(i) <- g.(k).(j); g.(k).(j) <- x
    done

           (* +-----------------------------------------+
              |  Changement d'origine dans un simplexe  |
              +-----------------------------------------+ *)

  (* vj <- vj-vi, vi <- -vi *)
  let fxsub v i =
    let n = Array.length(v) and p = Array.length(v.(0)) in
    for j=0 to n-1 do if j != i then
      for k=0 to p-1 do v.(j).(k) <- v.(j).(k) -- v.(i).(k) done
    done;
    vneg_in v.(i)

  (* vj <- vj-vi, vi <- -vi *)
  let exsub {v=v; g=g} i =
    fxsub v i;
    fxsub g i;
    let n = Array.length(g) in
    for j=0 to n-1 do
      if j != i then for k=0 to n-1 do	g.(k).(j) <- g.(k).(j) -- g.(k).(i) done
    done;
    for k=0 to n-1 do g.(k).(i) <- g.(i).(k) done;
    g.(i).(i) <- neg g.(i).(i)

                    (* +----------------------+
                       |  changement de base  |
                       +----------------------+ *)

  let fchg v m =
    let n = Array.length(v)        in
    let p = Array.length(v.(0))    in
    let w = Array.map Array.copy v in
    for i=0 to n-1 do
      for k=0 to p-1 do
	v.(i).(k) <- zero;
	for j=0 to n-1 do
	  v.(i).(k) <- v.(i).(k) ++ w.(j).(k)**m.(i).(j)
	done
      done
    done

  let echg {v=v; g=g} m =
    fchg v m;
    fchg g m;
    let n = Array.length(g)        in
    let h = Array.map Array.copy g in
    for i=0 to n-1 do
      for k=0 to n-1 do
	g.(k).(i) <- zero;
	for j=0 to n-1 do
	  g.(k).(i) <- g.(k).(i) ++ h.(k).(j)**m.(i).(j)
	done
      done
    done

                    (* +-----------------------+
                       |  Oprations diverses  |
                       +-----------------------+ *)


  (* transposition *)
  let transpose v =
    let n = Array.length(v)
    and p = Array.length(v.(0))        in
    let w = Array.make_matrix p n zero in
    for i=0 to p-1 do for j=0 to n-1 do w.(i).(j) <- v.(j).(i) done done;
    w

  (* dterminant et comatrice *)
  let rec iota a b = if a >= b then [] else a :: (iota (a+1) b)

  let rec determ vlist jlist = match vlist with
    | []  -> one
    | [v] -> v.(List.hd jlist)
    | v::vnext ->
	let rec calc jdeb jfin = match jfin with
	  | [] -> zero
	  | j::jnext -> if v.(j) <> zero
	    then (determ vnext (jdeb@jnext))**v.(j) -- (calc (jdeb@[j]) jnext)
	    else neg(calc (jdeb@[j]) jnext)
	in calc [] jlist

  let det v = determ (Array.to_list v) (iota 0 (Array.length v))

  let comatrice v =

    let n     = Array.length(v)  in
    let vlist = Array.to_list(v) in
    let jlist = iota 0 n         in

    let rec row vlist jdeb jfin s = match jfin with
      |	[] -> []
      | j::jnext -> let w = if s then determ vlist (jdeb@jnext)
                             	 else neg(determ vlist (jdeb@jnext)) in
                    w :: (row vlist (jdeb@[j]) jnext (not s))
    in

    let rec col vdeb vfin s = match vfin with
      | [] -> []
      | v::vnext -> let w = row (vdeb@vnext) [] jlist s in
                    (Array.of_list w) :: (col (vdeb@[v]) vnext (not s))
    in
    
    Array.of_list(col [] vlist true)

  (* matrice de Gram *)
  let gram v =
    let n = Array.length(v) in
    let g = Array.make_matrix n n zero in
    for i=0 to n-1 do
      for j=0 to i-1 do g.(i).(j) <- g.(j).(i)      done;
      for j=i to n-1 do g.(i).(j) <- ps v.(i) v.(j) done;
    done;
    g

  (* matrice identit *)
  let id n =
    let v = Array.make_matrix n n zero in
    for i=0 to n-1 do v.(i).(i) <- one done;
    v


                 (* +--------------------------------------+
                    |  numration des vecteurs primitifs  |
                    |        par norme 1 croissante        |
                    +--------------------------------------+ *)


  (* teste si un vecteur positif est primitif *)
  let primitif(v) =
    let n = Array.length(v) and d = ref(v.(0)) and i = ref(1) in
    while (!i < n) & (!d <> one) do d := gcd !d v.(!i); incr i done;
    !d = one

  (* numration des vecteurs positifs primitifs *)
  let rec next_pos(v) =
    let n = Array.length(v) in
    let k = v.(n-1) ++ one   in
    v.(n-1) <- zero;
    let i = ref(n-2)       in
    while (!i >= 0) & (v.(!i) = zero) do decr i done;
    v.(!i+1) <- k;
    if !i >= 0 then v.(!i) <- v.(!i) -- one;
    if not(primitif v) then next_pos(v)

  (* numration des vecteurs primitifs dont la *)
  (* premire coordonne non nulle est positive *)
  let vnext(v) =
    let i = ref(Array.length(v)-1) in
    while (!i >= 0) & (v.(!i) <= zero) do v.(!i) <- neg v.(!i); decr i done;
    let j = ref(!i-1) in
    while (!j >= 0) & (v.(!j) = zero) do decr j done;
    if !j >= 0 then v.(!i) <- neg v.(!i) else next_pos(v)


                     (* +--------------------+
                        |  Division centre  |
                        +--------------------+ *)

  let quomodc a b =
    let b1 = b >> 1 in
    let q,r = quomod (a++b1) b in (q,r--b1)

end

