(* +------------------------------------------------------------------------+
   |                                                                        |
   |                  Suite de Hamming : Rduction LLL                      |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

open Numerix
module M(E:Int_type) : sig

  (* transforme la famille fournie en famille rduite au sens de Lovascz *)
  val reduction : Linalg.M(E).euclide -> unit

end = struct

  module I = Infixes(E)
  module L = Linalg.M(E)
  open E
  open I
  open L

(*
  Soit (v0,v1,..,v(n-1)) une famille de vecteurs linairement indpendants.
  On note (w0,w1,..,w(n-1)) la famille de Schmidt associe :

  u(i) = v(i) + Sum(aij*v(j),j=0..i-1) avec u(i) orthogonal  v0,..,v(j-1).

  La famille (v0,..,v(n-1)) est reprsente dans ce module par deux
  matrices :

  v = (v.(i).(j)) = dcomposition dans la base canonique de Z^n

  g = (g.(i).(j)) = dcomposition dans (u0,..,u(n-1)) avec :

           g.(i).(0)      g.(i).(1)           g.(i).(i)
    v(i) = --------- u0 + --------- u1 + .. + --------- u(i)
           g.(0).(0)	  g.(1).(1)           g.(i).(i)

  et g.(0).(0) x g.(1).(1) x ... x g.(i).(i) = |u(i)|^2.
*)


          (* +----------------------------------------------------+
             |  Reprsentation euclidienne -> reprsentation LLL  |
             +----------------------------------------------------+ *)

    (*
      Transforme une matrice de Gram en dcomposition des v sur les u.
      Cette dcomposition s'obtient en chelonnant la matrice de Gram
      par l'algorithme de Gauss-Bareiss.   
    *)

    let schmidt g =
      let n = Array.length(g) in
      for i=0 to n-1 do
	for j=i+1 to n-1 do
	  for k=i+1 to n-1 do
	    let x = g.(i).(i)**g.(k).(j) -- g.(k).(i)**g.(i).(j) in
	    g.(k).(j) <- if i=0 then x else x//g.(i-1).(i-1)
	  done;
	  g.(i).(j) <- zero
	done;
      done

          (* +----------------------------------------------------+
             |  Reprsentation LLL -> reprsentation euclidienne  |
             +----------------------------------------------------+ *)
    (*
      Recalcule la matrice de Gram  partir des dcompositions des v sur les u.

      Formule :
      v(i) x v(j) = sum( g(i,k)*g(j,k)/g(k,k)g(k-1,k-1), k=0..min(i,j) )
      exprimentalement la somme arrte  p est un entier sur g(p,p)
      donc on peut conduire le calcul sans fractions comme dans Bareiss.
      (rdiger une preuve)
    *)
    let gram g =
      let n = Array.length(g) in

      for i=n-1 downto 1 do
	for j=i downto 1 do
	  let p = ref(zero) in
	  for k=0 to j-1 do
	    p := !p**g.(k).(k) ++ g.(i).(k)**g.(j).(k);
	    if k > 0 then p := !p//g.(k-1).(k-1)
	  done;
	  g.(i).(j) <- (!p ++ g.(i).(j))//g.(j-1).(j-1)
	done
      done;

      for i=0 to n-2 do
	for j=i+1 to n-1 do g.(i).(j) <- g.(j).(i) done
      done


             (* +-----------------------------------------------+
                |  rduction de v(j) par rapport  u(i), i < j  |
                +-----------------------------------------------+ *)

    let red {v=v; g=g} i j =
      let q,r = quomodc g.(j).(i) g.(i).(i) in
      if q <> zero then begin
         vpsub_in v.(j) q v.(i);
         vpsub_in g.(j) q g.(i)
      end


                  (* +------------------------------------+
                     |  change de v(i) et v(i-1), i > 0  |
                     +------------------------------------+ *)

    let swap {v=v; g=g} i =
      let n = Array.length(v) in
      let j = i-1            in

      (* transformation des lignes i et j *)
      let a = g.(i).(j)
      and b = if i=1 then one else g.(i-2).(i-2)
      and c = g.(i).(i)
      and d = g.(j).(j) in
      for k=j to n-1 do
	let x = g.(k).(j) and y = g.(k).(i) in
	g.(k).(j) <- (a**x ++ b**y)//d;
	g.(k).(i) <- (c**x -- a**y)//d
      done;

      (* change les colonnes i et j *)
      fswap v i j; fswap g i j


                           (* +------------------+
                              |  algorithme LLL  |
                              +------------------+ *)

    (*
      la sous-famille (v0,..,v(k-1)) est suppose LLL-rduite,
       rduit les vecteurs suivants.
    *)
    let rec lll ({v=v; g=g} as m) k = if k < Array.length(v) then begin

      (* rduit v(k) par rapport  u(k-1) *)
      red m (k-1) k;

      (* condition de Lovascz *)
      let x = if k=1 then g.(k).(k) else g.(k).(k)**g.(k-2).(k-2) in
      if (x ++ sqr(g.(k).(k-1)))*.4 >= sqr(g.(k-1).(k-1))*.3

      (* OK, continue la rduction *)
      then begin
	for i=0 to k-2 do red m i k done;
	lll m (k+1)
      end

      (* v(k) est trop court, change avec v(k-1) *)
      else begin
	swap m k;
	lll m (if k > 1 then k-1 else 1)
      end
    end


    let reduction ({v=v; g=g} as m) = schmidt g; lll m 1; gram g

end



