(* +------------------------------------------------------------------------+
   |                                                                        |
   |                Suite de Hamming : calcul de logarithmes                |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

open Numerix
module M(E:Int_type) : sig

  val ln235 : int -> E.t * E.t * E.t * E.t * E.t * E.t
  (* n -> (x,y,z,x',y',z') avec x <= 2^n*ln(2) <= x'
                                y <= 2^n*ln(3) <= y'
                                z <= 2^n*ln(5) <= z' *)
end = struct

    (* on calcule

       [ x ]   [   1  -27   18 ]   [ ln(2) ]     [ argth(a) ]
       [ y ] = [  38   -2  -15 ] * [ ln(3) ] = 2*[ argth(b) ]
       [ z ]   [ -15    8    1 ]   [ ln(5) ]     [ argth(c) ]
       
       avec
       a = (2^1*5^18 - 3^27)     / (2^1*5^18 + 3^27)     ~ 25e-5
       b = (2^38     - 3^2*5^15) / (2^38     + 3^2*5^15) ~ 40e-5
       c = (3^8*5^1  - 2^15)     / (3^8*5^1  - 2^15)     ~ 56e-5
       
       et on en dduit
       
       [ ln(2) ]   [ 118  171  441 ]   [ x ]
       [ ln(3) ] = [ 187  271  699 ] * [ y ]
       [ ln(5) ]   [ 274  397 1024 ]   [ z ]
       
       Les coefficients 1,-27,18,... ont t trouvs par recherche
       systmatique de nombres de Hamming conscutifs proches de faon
        avoir des rapports proches de 1. a doit pouvoir tre automatis
       en appliquant LLL sur des approximations en flottant de ln(2),
       ln(3) et ln(5).
    *)

  module I = Infixes(E)
  open E
  open I
  let zero  = of_int(0)
  let un    = of_int(1)
  let deux  = of_int(2)
  let trois = of_int(3)
  let cinq  = of_int(5)

           (* +---------------------------------------------------+
              |  encadrement de argth((a-b)/(a+b))  2^(-n) prs  |
              +---------------------------------------------------+ *)

  let argth a b n =
    let u  = a--b    and v  = a++b    in
    let u2 = sqr(u) and v2 = sqr(v) in
    let v2u2 = v2--u2 in

    (* calcule le dbut de la srie entire et arrte quand le reste *)
    (* est assez petit. k = 2i+1, uk = u^k, vk = v^k,                *)
    (* p/(q*vk) = somme jusqu' uk/kvk.                              *)
    let rec calc k uk vk p q =
      let x = uk**u2         in
      let y = vk**v2         in
      let l = k+2          in
      let r = p**v2*.l ++ q**x in
      let s = q*.l          in
      if nbits(x) - nbits(vk) - nbits(v2u2) > -n then calc l x y r s
      else let s  = s**vk        in
	   let r' = r -- p**u2*.l in
	   let s' = s**v2u2      in
	   (r << n)//(s**v2), ((r' << n) ++ s'--un)//s'

    in calc 1 u v u un

              (* +--------------------------------------------+
                 |  calcule ln(2), ln(3) et ln(5) sur n bits  |
                 +--------------------------------------------+ *)

  let ln235 n =

    let prec = n + 17 in (* prend 16 bits de rserve + 1 pour 2*argth *)
    let x,x' = argth (deux ** (cinq^^18)) (trois^^27)            prec
    and y,y' = argth (deux^^38)        ((trois^^2) ** (cinq^^15))  prec
    and z,z' = argth ((trois^^8) ** cinq) (deux^^15)             prec
    and deux_16 = (un << 16) -- un in
    ( (* minorants *)
      (x*.118  ++ y*.171  ++ z*.441)  >> 16,
      (x*.187  ++ y*.271  ++ z*.699)  >> 16,
      (x*.274  ++ y*.397  ++ z*.1024) >> 16,
      (* majorants *)
      (x'*.118 ++ y'*.171 ++ z'*.441  ++ deux_16) >> 16, 
      (x'*.187 ++ y'*.271 ++ z'*.699  ++ deux_16) >> 16,
      (x'*.274 ++ y'*.397 ++ z'*.1024 ++ deux_16) >> 16
    )

end

