(* +------------------------------------------------------------------------+
   |                                                                        |
   |               Suite de Hamming : Dnombrement d'un simplexe            |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

open Printf
open Numerix

module M(E:Int_type) : sig

  (* reprsentation abstraite d'un simplexe *)
  type simplexe
  val nbcones : simplexe -> int

  (* simplexe a0*x0 + .. + a(n-1)*x(n-1) <= k, x(i) >= 0 *)
  val make : (Linalg.M(E).euclide -> unit) -> Linalg.M(E).vecteur -> simplexe

  (* volume d'un simplexe et drives partielles en fonction de k *)
  val volume : simplexe -> E.t -> E.t
  val derive : simplexe -> E.t -> E.t -> Linalg.M(E).vecteur

  (* volume approch et drive par rapport  k *)
  val approx : simplexe -> E.t -> (E.t*E.t*E.t) 

end = struct

  module I = Infixes(E)
  module L = Linalg.M(E)
  module C = Cone.M(E)
  open E
  open I
  open L
  open C

  let zero  = of_int(0)
  and un    = of_int(1)
  and deux  = of_int(2)
  and trois = of_int(3)
  and quatre= of_int(4)
  and cinq  = of_int(5)
  and six   = of_int(6)
  and douze = of_int(12)

  
  (* Algorithmes :

     Pour calculer le nombre de points d'un simplexe on value
     en (1,1,1) sa srie gnratrice obtenue comme somme de
     fractions associes aux cnes unimodulaires de chaque coin.

     Comme (1,1,1) est un ple de chaque fraction on forme un DL
     de chaque fraction et on calcule la somme des termes constants.

     Une fraction est de la forme :


                            x^u1*y^u2*z^u3
          ------------------------------------------------------
          (1-x^a1*y^a2*z^a3)(1-x^b1*y^b2*z^b3)(1-x^c1*y^c2*z^c3)

     On pose x=e^it, y=e^jt, z=e^kt o (i,j,k) est un vecteur
     qui n'est orthogonal  aucun des vecteurs (a1,a2,a3), (b1,b2,b3)
     et (c1,c2,c3) ce qui donne la fraction :

                                 e^ut
                       ------------------------
                       (1-e^at)(1-e^bt)(1-e^ct)

     dont le terme constant du DL en t=0 est :

                     (s1 - 2u) (s2 -2u*s1 + 2u^2)
                     ---------------------------- 
                               24abc

     avec s1 = a+b+c et s2 = ab+ac+bc.

     Pour viter les calculs rationnels on met 1/24 et facteur
     et on calcule (s1-2u)(s2-2us1+2u^2)/abc en virgule fixe
     avec 32 bits aprs la virgule. Le rsultat, aprs arrondi,
     est exact tant que le nombre de cnes unimodulaires est
     infrieur  24*2^31 ~ 51 milliards.


     Une fois le n-me point de Hamming isol entre deux ttradres
     rationnels de fonctions gnratrices f et g, on dtermine les
     coordonnes de ce point en valuant en (1,1,1) les quantits :

                     d(f-g)  d(f-g)    d(f-g)
                     ------, ------ et ------
                       dx      dy        dz

     car f-g = x^h2*y^h3*z^h5 o (h2,h3,h5) est le point cherch.
     Les drives partielles de f-g sont values de la mme manire
     par addition des termes constants des DL en 0 des drives
     partielles de chaque fraction avec x=e^it, y=e^jt, z=e^kt.
     Une fraction composant f-g est de la forme :

                    x^u1*y^u2*z^u3 - x^v1*y^v2*z^v3
          ------------------------------------------------------
          (1-x^a1*y^a2*z^a3)(1-x^b1*y^b2*z^b3)(1-x^c1*y^c2*z^c3)

     ce qui donne les termes constants suivants pour la drive
     par rapport  x :

           /    u1(s1 - 2(u1-i))(s2 - 2(u1-i)s1 + 2(u1-i)^2)            \
       1   |  - v1(s1 - 2(v1-i))(s2 - 2(v1-i)s1 + 2(v1-i)^2)            |
     ----- |  + (u-v)(ua + va - sa1)(sa2 - (ua+va)sa1 + (ua^2+va^2))/a  |
     24abc |  + (u-v)(ub + vb - sb1)(sb2 - (ub+vb)sb1 + (ub^2+vb^2))/b  |
           \  + (u-v)(uc + vc - sc1)(sc2 - (uc+vc)sc1 + (uc^2+vc^2))/c  /

     avec sa1 = a + a + b + c,  sa2 = a^2 + 2ab + 2ac + bc
          sb1 = a + b + b + c,  sb2 = b^2 + 2ab + ac  + 2bc
          sc1 = a + b + c + c,  sc2 = c^2 + ab  + 2ac + 2bc

     et   ua = u + a - i,  ub = u + b - j,  uc = u + c - k
          va = v + a - i,  vb = v + b - j,  vc = v + c - k

     En calculant ces quantits avec 32 bits aprs la virgule on
     obtient un arrondi correct tant que le nombre de cnes
     unimodulaires est infrieur  24*2^27 ~ 3.2 milliards.

  *)

  let bits  = 32
  let denom = (of_int 24) << bits

  (* fraction prte  calculer *)
  type fraction = {
    s1  : E.t;   (* a+b+c        *)
    s2  : E.t;   (* ab + ac + bc *)
    abc : E.t    (* abc          *)
  }

  (* reprsentation d'un simplexe : coefficients, cnes *)
  (* unimodulaires et exposants d'approche              *)

  type simplexe = {
    coef   : vecteur;             (* coefficients de la face oblique *)
    exp    : vecteur;             (* exposants d'approche            *)
    cones  : cone list array;     (* cnes regroups par sommet      *)
    fracts : fraction list array; (* fractions prcalcules          *)
    f0     : E.t                  (* terme constant (coin origine)   *)
  } 

  let nbcones s =
    let res = ref(1) in
    for i=0 to Array.length(s.cones)-1 do
      res := !res + List.length(s.cones.(i))
    done;
    !res

        (* +----------------------------------------------+
           |  Division centre / arrondie suprieurement  |
           +----------------------------------------------+ *)

  let quoc a b = let q,_ = quomodc a b in q
  let quoh a b = quo (a++b--un) b


         (* +--------------------------------------------+
            |  Dcompose un coin en cnes unimodulaires  |
            +--------------------------------------------+ *)

  (* pour viter les cnes de dimesion < 3 on dcompose le cne dual *)
  let dcompose reduction coef i =
    let n = Array.length(coef)         in
    let f = id(n) in
    f.(i) <- vneg coef;
    barvinok reduction [C.make 1 f]

            (* +--------------------------------------+
               |  Recherche des exposants d'approche  |
               +--------------------------------------+ *)

  (* on numre les petits vecteurs jusqu' en trouver un *)
  (* orthogonal  aucun des vecteurs portant les cnes    *)

  let exposants cones =

    (* met en liste tous les vecteurs apparaissant aux dnominateurs *)
    let l = ref [] in
    Array.iter (List.iter (fun c -> Array.iter (fun x -> l := x :: !l) c.g)) cones;

    (* cherche un vecteur orthogonal  aucun des lments de l *)
    let e = Array.make (Array.length cones) un in
    let rec search liste = match liste with
      | [] -> ()
      | x::suite when ps e x <> zero -> search suite
      | _ -> vnext e; search !l
    in
    search !l; e


                  (* +---------------------------+
                     |  prcalcule une fraction  |
                     +---------------------------+ *)

    let fract e c =
      let a = Array.map (ps e) c.g in
      {s1  = a.(0) ++ a.(1) ++ a.(2);
       s2  = a.(0)**a.(1) ++ a.(0)**a.(2) ++ a.(1)**a.(2);
       abc = a.(0)**a.(1)**a.(2)}


                   (* +-------------------------+
                      |  Constitue un simplexe  |
                      +-------------------------+ *)

  let make reduction coef =

    let n = Array.length(coef) in
    if n <> 3 then failwith "volume non implment en dimension <> 3";

    let cones  = Array.init n (dcompose reduction coef)      in
    let exp    = exposants cones                              in
    let fracts = Array.map (List.map (fract exp)) cones       in
    let a      = exp.(0) and b = exp.(1) and c = exp.(2)      in
    let f0     = quoc ((a++b++c)**(a**b++a**c++b**c) << bits) (a**b**c) in
    {coef = coef; cones = cones; exp = exp; fracts = fracts; f0 = f0}


                (* +-------------------------------+
                   |  origine d'un cne translat  |
                   +-------------------------------+ *)

    let origine coef cone k i =
      let x = Array.map (fun v -> quoh (k**v.(i)) coef.(i)) cone.f in
      vcomb x cone.g
	
        (* +-----------------------------------------------+
           |  volume a(0)*x(0) + ... + a(n-1)*x(n-1) <= k  |
           +-----------------------------------------------+ *)

  let volume s k =
    let n = Array.length(s.exp) in
    let l = ref s.f0            in

    (* i = n du coin, c = cne, f = fraction *)
    let job i c f =
      let u = ps s.exp (origine s.coef c k i)               in
      let x =  (f.s1 -- u*.2)**(f.s2 ++ u**(u--f.s1)*.2) << bits in
      l := !l ++ (quoc x f.abc)*.c.s
    in

    for i=0 to n-1 do List.iter2 (job i) s.cones.(i) s.fracts.(i) done;
    quoc !l denom

                    (* +-----------------------+
                       |  drives partielles  |
                       +-----------------------+ *)

  let derive s k k' =
    let n = Array.length(s.exp) in
    let l = Array.make n zero   in
    let t = ref zero            in

    (* i = n du coin, c = cne, f = fraction *)
    let job i c f =

      let o = origine s.coef c k i and o' = origine s.coef c k' i in
      let u = ps s.exp o           and v  = ps s.exp o'           in

      (* j = n drive partielle *)
      for j = 0 to n-1 do

	let u = u -- s.exp.(j) and v = v -- s.exp.(j)  in
	t := (  o.(j) **(f.s1 -- u*.2)**(f.s2 ++ u**(u -- f.s1)*.2)
              -- o'.(j)**(f.s1 -- v*.2)**(f.s2 ++ v**(v -- f.s1)*.2)) << bits;

	(* k = n du facteur au dnominateur *)
	let du = u -- v in
	if du <> zero then for k = 0 to n-1 do
	  let a  = ps s.exp c.g.(k)                in
	  let u  = u ++ a    and v = v ++ a
  	  and s1 = f.s1 ++ a and s2 = f.s2 ++ a**f.s1 in
	  let su = u ++ v                           in
	  let x  = (su -- s1)**(sqr(u) ++ sqr(v) -- su**s1 ++ s2) in
	  t := !t ++ (quoc (c.g.(k).(j)**du**x << bits) a)
	done;

	l.(j) <- l.(j) ++ (quoc !t f.abc)*.c.s

      done
    in

    for i=0 to n-1 do List.iter2 (job i) s.cones.(i) s.fracts.(i) done;
    for j=0 to n-1 do l.(j) <- quoc l.(j) denom done;
    l


                          (* +---------------------+
                             |  Formule approche  |
                             +---------------------+ *)

  (* retourne v(a,b,c,k) et v'(a,b,c,k) (numrateur et dnominateur spars) *)
  (* formules dduites du DSE de 1/(1-t)(1-t^a)(1-t^b)(1-t^c)                  *)
  let approx s k =
    let a  = s.coef.(0) and b = s.coef.(1) and c = s.coef.(2) in
    let a' = a--un      and b' = b--un     and c' = c--un     in
    let s  = a' ++ b' ++ c'
    and p  = a'**b' ++ a'**c' ++ b'**c'
    and d  = douze**a**b**c in
    let v  = (s++p)**(s++deux)
	     ++ (k++un)**(deux**(s**(s++deux)++p)
		       ++ (k++deux)**(six**s ++ quatre**(k++trois)))
    and v' = (s++un)**(s++quatre)++p ++ six**(k++un)**(k++s++trois)
    in (v//(deux**d), v',d)

end
